# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

INCOMING_MESSAGES = 'IncomingMessages'
INCOMING_BYTES = 'IncomingBytes'
OUTGOING_MESSAGES = 'OutgoingMessages'
OUTGOING_BYTES = 'OutgoingBytes'
